/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// File       : UHistogram.cc
// Creation   : 06/1996
// Collab     : N.OLIVER
// Goal       : defines the THistogram class, see "UHistogram.h"
//
// Tab        : none
//  Nuria Oliver 1997, nuria@media.mit.edu

#include "stdafx.h"
// unit include
#include "UHistogram.h"

// project includes
#include "UObjAlloc.h"

// system includes
#include "string.h"


//
//    Init
//

//		Assumes that the encoding is the right one
ErrorNumber THistogram::Init (DWORD reso, BOOL prob)
{
	ErrorNumber     err = kErrNoErr;
	
	// setup data members
	m_pValues			= (LPBYTE)NULL;
	m_pLongValues		= (LPDWORD)NULL;
	m_pProbValues = (float*)NULL;
	m_Reso			= reso;
	m_pColorModel		= (TGaussianf*)NULL;
	m_fMaxDist		= (float)0.0;
	m_bRobust			= FALSE;
	m_bProb		=	prob;
	
	// test if we can manage such pixel encoding
	// TO BE DONE
	if (m_Reso == 0)
		// enforce a reasonable value
		m_Reso = 256;
	
	// allocate memory for histogram
	m_pLongValues = new DWORD[m_Reso * m_Reso];
	
	if (m_pLongValues == (LPDWORD)NULL)
		err = kErrMemFull;
	else
	{
		m_pValues = (LPBYTE)m_pLongValues;	  
		// allocate the ColorModel
		err = Allocate(m_pColorModel);
		if ( err == kErrNoErr ) {
			if ( m_bProb == TRUE ) 
				m_pProbValues = new float[m_Reso * m_Reso];
		}
	}
	
	
	// error handling
	if (err != kErrNoErr)
		this->Dispose ();
	
	return err;
}





//
//    Dispose
//

void THistogram::Dispose ()
{
	// dispose the colormodel
	if (m_pColorModel != (TGaussianf*)NULL);
	{
		delete m_pColorModel;
		m_pColorModel = (TGaussianf*)NULL;
	}
	
	// dispose memory for values
	if (m_pLongValues != (LPDWORD)NULL)
	{
		delete[] m_pLongValues;
		m_pLongValues = (LPDWORD)NULL;
	}
	// dispose memory for prob values
	if (m_pProbValues != (float *)NULL)
	{
		delete[] m_pProbValues;
		m_pProbValues = (float *)NULL;
	}
}




//
//    Compute
//

ErrorNumber THistogram::Compute (	VDib*	bitmap, 
								 BOOL		robust,
								 BYTE		threshold, 
								 CRect*		sourceRect, 
								 float		maxDist )
{
	ErrorNumber	err = kErrNoErr;
	LPBYTE		linePtr;
	LPBYTE		pixelPtr;
	CRect			rect;
	int			pixelSize = bitmap->PixelSize()/8; // Pixel size in bytes
	DWORD			x, y, ylimit, newMax;
	DWORD         r, g, b, lum, multip = m_Reso - 1;
	DWORD         lineOffset;
	DWORD         maxCard;
	
	// Check for encoding in the future
	
	// get bitmap region of interest
	rect = (sourceRect!=(CRect*)NULL) ? *sourceRect : bitmap->GetValidRect();
	
	// get base address of bitmap memory
	linePtr = (LPBYTE)(bitmap->GetBits ());
	
	// get accelerator for next line
	lineOffset = bitmap->LineOffset();
	
	// go to first valid line
	linePtr += rect.top * lineOffset;
	
	// go to first valid pixel of that line
	linePtr += rect.left * pixelSize;
	
	if (robust)
	{
		// Reset the model
		m_pColorModel->setZero ();
		
		// cycle for lines
		for (y = rect.bottom - rect.top; y > 0; y--)
		{
			pixelPtr = linePtr;
			
			// cycle for pixels
			for (x = rect.right - rect.left; x > 0; x--)
			{
				// get 3 pixel components: BGRX
				b = *pixelPtr;
				pixelPtr++;
				g = *pixelPtr;
				pixelPtr++;
				r = *pixelPtr;
				pixelPtr+=2;
				
				// compute pixel luminance
				lum = r+g+b;
				
				// add vector to color model
				if (lum)
					m_pColorModel->addSample
					((TVGAUSSTYPE)r/lum, (TVGAUSSTYPE)g/lum);
				else
					m_pColorModel->addSample ((TVGAUSSTYPE)0, (TVGAUSSTYPE)0);
			}
			linePtr += lineOffset;
		}
		
		// compute mean and covariance of model
		m_pColorModel->getGaussianInv ();
		
		// build the histogram
		err = this->BuildFromStats (maxDist, threshold);
		
		m_bRobust = (err == kErrNoErr) ? TRUE : FALSE;
	}
	else
	{
		m_bRobust = FALSE;
		
		// init classes max cardinal
		maxCard = 0;
		
		// Init histogram values
		memset (m_pLongValues, 0, m_Reso*m_Reso*sizeof(DWORD));
		
		// cycle for lines
		for (y = rect.bottom - rect.top; y > 0; y--)
		{
			pixelPtr = linePtr;
			
			// cycle for pixels
			for (x = rect.right - rect.left; x > 0; x--)
			{
				// get 3 pixel components: BGRX
				b = *pixelPtr;
				pixelPtr++;
				g = *pixelPtr;
				pixelPtr++;
				r = *pixelPtr;
				pixelPtr+=2;
				
				// compute pixel luminance
				lum = r+g+b;
				
				// increment histogram for that pixel value
				if (lum)
					newMax = (m_pLongValues[(r*multip/lum)*m_Reso + g*multip/lum] += 1);
				else
					newMax = (m_pLongValues[0] += 1);
				
				if (newMax > maxCard)
					maxCard = newMax;
			}
			linePtr += lineOffset;
		}
		
		// scale the histogram between 0 and 255
		// build the histogram
		if (threshold == 0)
			for (x = 0; x < m_Reso; x++)
			{
				ylimit = m_Reso-x;
				
				for (y = 0; y < ylimit; y++)
					m_pValues[x*m_Reso+y] = (unsigned char) ((BYTE)m_pLongValues[x*m_Reso+y] * 255 / maxCard);
			}
			else
				for (x = 0; x < m_Reso; x++)
				{
					ylimit = m_Reso-x;
					
					for (y = 0; y < ylimit; y++)
						m_pValues[x*m_Reso+y] =
						(m_pLongValues[x*m_Reso+y] * 255 / maxCard) >= threshold ? 255: 0;
				}
	} //else
	
	return err;
}




//
//    Apply
//

ErrorNumber THistogram::Apply (	TBitMap* bitmap, 
							   TBitMap* res, 
							   CRect* sourceRect,
							   CRect* destRect)
{
    ErrorNumber		err = kErrNoErr;
    CRect			sRect = (sourceRect!=(CRect*)NULL) ? *sourceRect : bitmap->GetValidRect();
    CRect			dRect = (destRect!=(CRect*)NULL) ? *destRect : res->GetValidRect();
    DWORD			x, y;
    BYTE			*slPtr, *dlPtr, *spPtr, *dpPtr;
    DWORD			sOffs, dOffs;
    DWORD			r, g, b, lum;
    DWORD			multip = m_Reso - 1;
    
	// Check for correct pixel sizes: origin must be RGBA (32 bits)
	// and destination must be G (8 bit)
    if ((bitmap->PixelSize() != 32) ||
		(res->PixelSize() != 8))
		err = kErrCantHandleEncoding;
    else
	{
		// test if the validrects have the same size
		if (sRect != dRect ) {
			err = kErrWrongSize;
		} else {
			// get source ans dest line byte offset
			sOffs = bitmap->LineOffset();
			dOffs = res->LineOffset();
			
			// setup line ptrs
			slPtr = (LPBYTE)(bitmap->GetBuffer ()) + sRect.top * sOffs + sRect.left * 4;
			dlPtr = (LPBYTE)(res->GetBuffer ()) + dRect.top * dOffs + dRect.left;
			
			// cycle for lines
			for (y = sRect.bottom - sRect.top; y>0; y--)
			{
				spPtr = slPtr;
				dpPtr = dlPtr;
				
				// cycle for pixels in line
				for (x = sRect.right - sRect.left; x>0; x--)
				{
					// get 3 pixel components: BGRX
					b = *spPtr;
					spPtr++;
					g = *spPtr;
					spPtr++;
					r = *spPtr;
					spPtr+=2;
					
					// compute pixel luminance
					lum = r+g+b;
					
					// set destination value
					if (lum)
						*dpPtr = m_pValues[r*multip/lum*m_Reso + g*multip/lum];
					else
						*dpPtr = m_pValues[0];
					
					dpPtr++;       
				} //for
				
				// go to next line
				slPtr += sOffs;
				dlPtr += dOffs;
			} //for
		} //else
	}//else
    return err;
}


ErrorNumber     THistogram::ApplyProbMap  (TBitMap* bitmap, 
										   TBitMap* res, 
										   VDib*	resProb,
										   CRect* sourceRect,
										   CRect* destRect )
{
    ErrorNumber   err = kErrNoErr;
    CRect         sRect = (sourceRect!=(CRect*)NULL) ? *sourceRect : bitmap->GetValidRect();
    CRect         dRect = (destRect!=(CRect*)NULL) ? *destRect : res->GetValidRect();
    long        x, y;
    BYTE         *slPtr, *dlPtr, *spPtr, *dpPtr;
    float         *dProbPtr;
    long        sOffs, dOffs;
    long        r, g, b, lum, index;
    long        multip = m_Reso - 1;
    
    if ((bitmap->PixelSize() != 32) ||
		(resProb->PixelSize() != 32) ||
		(res->PixelSize() != 8) ||
		( m_bProb == false ) )
	{
		err = kErrCantHandleEncoding;
	}
    else
	{
		// test if the validrects have the same size
		if (sRect != dRect) { 
			err = kErrWrongSize;
		} else {
			// get source ans dest line byte offset
			sOffs = bitmap->LineOffset();
			dOffs = res->LineOffset();
			
			// setup line ptrs
			slPtr = (LPBYTE)(bitmap->GetBuffer ()) + sRect.top * sOffs + sRect.left * 4;
			dlPtr = (LPBYTE)(res->GetBuffer ()) + dRect.top * dOffs + dRect.left;
			dProbPtr = (float *)(resProb->GetBits ()) + dRect.top * dOffs + dRect.left;
			
			// cycle for lines
			for (y = sRect.bottom - sRect.top; y>0; y--)
			{
				spPtr = slPtr;
				dpPtr = dlPtr;
				
				// cycle for pixels in line
				for (x = sRect.right - sRect.left; x>0; x-- )
				{
					// get 3 pixel components: BGRX
					b = *spPtr;
					spPtr++;
					g = *spPtr;
					spPtr++;
					r = *spPtr;
					spPtr+=2;
					
					// compute pixel luminance
					lum = r+g+b;
					
					// set destination value
					if (lum)
					{
						index = r*multip/lum*m_Reso + g*multip/lum;
						*dpPtr = m_pValues[index];
						*dProbPtr = m_pProbValues[index];
					}
					else
					{
						*dpPtr = m_pValues[0];
						*dProbPtr = m_pProbValues[0];
					}
					dpPtr++;       
					dProbPtr++;
				} //for
				
				// go to next line
				slPtr += sOffs;
				dlPtr += dOffs;
			} //for
			
		} //else
	}//else
    return err;
}


//
//    BuildFromStatArray
//  
// 1.array has the array with the POSITIVE gaussians followed by the NEGATIVE
//   gaussians
// 2.weights is an array of floating point weights to weight the distance
//   of each gaussian
// 3.nbEntry=total number of entries
// 4.nbPositive=total number of positive gaussians
// 5.maxDist=maxDist for binarizing the histogram
//

ErrorNumber THistogram::BuildFromStatArray (TGaussianf** array,
                                            float* weights,
                                            int nbEntry,
                                            int nbPositive,
                                            float maxDist)
{
	ErrorNumber       err = kErrNoErr;
	unsigned long            x, y, ylimit;
	float             dist;
	float             min;
	int               minInd, j;
	
	if ((nbEntry < 2) || (nbPositive < 1) || (nbEntry - nbPositive < 0))
		err = kErrBadParams;
	else
	{
		if ( m_bProb == FALSE )
		{
			if (weights != (float*)NULL)
			{
				// build the histogram
				for (x = 0; x < m_Reso; x++)
				{
					ylimit = m_Reso-x;
					
					for (y = 0; y < ylimit; y++)
					{
						min     = maxDist;
						minInd  = nbEntry;
						
						for (j = nbEntry-1; j >= 0; j--)
						{
							// Compute mininum distance to all the models, divided
							// by the weight->greater weight smaller distance
							// and it needs to be LESS than the MAXDIST provided
							dist = (array[j])->getMahalanobisSquare ((TVGAUSSTYPE)x/m_Reso,
								(TVGAUSSTYPE)y/m_Reso) /
								weights[j];
							if (dist < min)
							{
								min = dist;
								minInd = j;
							}
						}
						// if minimum distance is to a positive gaussian, then assign
						// 255, if not, assign 0
						m_pValues[x*m_Reso+y] = (minInd < nbPositive) ? 255 : 0;
					}
				}
			} else // weights != NULL
			{
				// build the histogram
				for (x = 0; x < m_Reso; x++)
				{
					ylimit = m_Reso-x;
					
					for (y = 0; y < ylimit; y++)
					{
						min     = maxDist;
						minInd  = nbEntry;
						
						for (j = nbEntry-1; j >= 0; j--)
						{
							dist = (array[j])->getMahalanobisSquare ((TVGAUSSTYPE)x/m_Reso,
								(TVGAUSSTYPE)y/m_Reso) /
								maxDist;
							if (dist < min)
							{
								min = dist;
								minInd = j;
							}
						}
						m_pValues[x*m_Reso+y] = (minInd < nbPositive) ? 255 : 0;
					}
				}
			}
		}
	} // else
	return err;
}


//
//    BuildFromMixModels
//  
// 1.array has the array with the POSITIVE gaussians followed by the NEGATIVE
//   gaussians
// 2.weights is an array of floating point weights to weight the distance
//   of each gaussian
// 3.nbEntry=total number of entries
// 4.nbPositive=total number of positive gaussians
// 5.maxDist=maxDist for binarizing the histogram
//

ErrorNumber THistogram::BuildFromMixModels (TGaussianf* array,
                                            float* weights,
                                            int nbEntry,
                                            float maxDist)
{
	ErrorNumber       err = kErrNoErr;
	unsigned long            x, y, ylimit;
	float             dist;
	float             min;
	int               minInd, j;
	
	
	if (weights != (float*)NULL)
    {
		//for ( int i=0;i<nbEntry;i++)
		//	{
		//  cout << "Array in BuildFromMixModels " << (array[i]);
		//  cout << " weights " << weights[i] << endl;
		//}
		// build the histogram
		for (x = 0; x < m_Reso; x++)
		{
			ylimit = m_Reso-x;
			
			for (y = 0; y < ylimit; y++)
			{
				min     = maxDist;
				minInd  = nbEntry;
				dist = 0;
				for (j = nbEntry-1; j >= 0; j--)
				{
					// Compute mininum distance to all the models, divided
					// by the weight->greater weight smaller distance
					// and it needs to be LESS than the MAXDIST provided
					dist += 
						(array[j]).getMahalanobisSquare ((TVGAUSSTYPE)x/m_Reso,
						(TVGAUSSTYPE)y/m_Reso)*
						weights[j];
				}
				if (dist < min)
				{
					min = dist;
					minInd = j;
				}
				// if minimum distance is to a positive gaussian, then assign
				// 255, if not, assign 0
				m_pValues[x*m_Reso+y] = (minInd < nbEntry) ? 255 : 0;
			}
		}
    } else
	{
		//for ( int i=0;i<nbEntry;i++)
		// {
		//  cout << "Array in BuildFromMixModels " << (array[i]);
		//}
		// build the histogram
		for (x = 0; x < m_Reso; x++)
		{
			ylimit = m_Reso-x;
			
			for (y = 0; y < ylimit; y++)
			{
				min     = maxDist;
				minInd  = nbEntry;
				dist = 0;
				for (j = nbEntry-1; j >= 0; j--)
				{
					dist += (array[j]).getMahalanobisSquare ((TVGAUSSTYPE)x/m_Reso, (TVGAUSSTYPE)y/m_Reso)*1/nbEntry;
					if (dist < min)
					{
						min = dist;
						minInd = j;
					}
				}
				m_pValues[x*m_Reso+y] = (minInd < nbEntry) ? 255 : 0;
			}
		}
	} // else
	return err;
}

//
//  BuildFrom1MixModel
//
//  Computes the probability of a mixture model and it thresholds
// the histogram based on this probability and the min probability
// required

ErrorNumber THistogram::BuildFrom1MixModel ( TGaussianf* posGauss,
											float* posW,
											int nbPosEntry,
											float maxPosDist,
											BOOL normalize )
{
	ErrorNumber       err = kErrNoErr;
	DWORD            x, y, ylimit;
	float             posDist, totalProb=0.0;
	// float             min;
	//int               minInd, i, j;
	int j;
	float             Z = 1;
	
	if ( normalize == TRUE ) {
		// compute the Z = normalization function
		// build the histogram
		for (x = 0; x < m_Reso; x++)
		{
			ylimit = m_Reso-x;
			
			for (y = 0; y < ylimit; y++)
			{
				for (j = nbPosEntry-1; j >= 0; j--)
				{
					// Compute mininum distance to all the models, divided
					// by the weight->greater weight smaller distance
					// and it needs to be LESS than the MAXDIST provided
					totalProb += 
						(posGauss[j]).getProbability ((TVGAUSSTYPE)x/m_Reso,
						(TVGAUSSTYPE)y/m_Reso)*posW[j];
				}
				//if ( posDist > 0 )
				//  posDist = -log(posDist);
				
				// if minimum distance is to a positive gaussian, then assign
				// 255, if not, assign 0
				//	  fValues[x*m_Reso+y] = ( (posDist > negDist) && (posDist < maxPosDist )) ? 255 : 0;
				//if ( posDist > maxDist )
				//  maxDist = posDist;
			} // for y
		} // for x
	}
	if ( totalProb > 0 ) 
		Z = totalProb/m_Reso;
	if ( !m_bProb ) {
		// build the histogram
		for (x = 0; x < m_Reso; x++)
		{
			ylimit = m_Reso-x;
			
			for (y = 0; y < ylimit; y++)
			{
				posDist = 0;
				for (j = nbPosEntry-1; j >= 0; j--)
				{
					// Compute mininum distance to all the models, divided
					// by the weight->greater weight smaller distance
					// and it needs to be LESS than the MAXDIST provided
					posDist += 
						(posGauss[j]).getProbability ((TVGAUSSTYPE)x/m_Reso,
						(TVGAUSSTYPE)y/m_Reso)*posW[j];
				}
				//if ( posDist > 0 )
				//  posDist = -log(posDist);
				
				// if minimum distance is to a positive gaussian, then assign
				// 255, if not, assign 0
				//	  fValues[x*m_Reso+y] = ( (posDist > negDist) && (posDist < maxPosDist )) ? 255 : 0;
				//if ( posDist > maxDist )
				//  maxDist = posDist;
				posDist /= Z;
				m_pValues[x*m_Reso+y] = ( posDist > maxPosDist ) ? (unsigned char) (255*Z*posDist) : 0;
			} // for y
		} // for x
	} else { //m_bProb=1
		for (x = 0; x < m_Reso; x++)
		{
			ylimit = m_Reso-x;
			
			for (y = 0; y < ylimit; y++)
			{
				posDist = 0;
				for (j = nbPosEntry-1; j >= 0; j--)
				{
					// Compute mininum distance to all the models, divided
					// by the weight->greater weight smaller distance
					// and it needs to be LESS than the MAXDIST provided
					posDist += 
						(posGauss[j]).getProbability ((TVGAUSSTYPE)x/m_Reso,
						(TVGAUSSTYPE)y/m_Reso)*posW[j];
				}
				//if ( posDist > 0 )
				//  posDist = -log(posDist);
				
				// if minimum distance is to a positive gaussian, then assign
				// 255, if not, assign 0
				//	  fValues[x*m_Reso+y] = ( (posDist > negDist) && (posDist < maxPosDist )) ? 255 : 0;
				//if ( posDist > maxDist )
				//  maxDist = posDist;
				posDist /= Z;
				m_pProbValues[x*m_Reso+y]=posDist;
				m_pValues[x*m_Reso+y] = ( posDist > maxPosDist ) ? (unsigned char) (255*posDist*Z) : 0;
			} // for y
		} // for x
	}
	return err;
}




//
//    BuildFrom2MixModels->computes probabilities for each model
//  
// 1.array has the array with the POSITIVE gaussians followed by the NEGATIVE
//   gaussians
// 2.weights is an array of floating point weights to weight the probability
//   of each gaussian
// 3.nbEntry=total number of entries
// 4.nbPositive=total number of positive gaussians
// 5.maxDist=maxDist for binarizing the histogram


ErrorNumber THistogram::BuildFrom2MixModels ( TGaussianf* posGauss,
											 float* posW,
											 int nbPosEntry,
											 float maxPosDist,
											 TGaussianf* negGauss,
											 float* negW,
											 int nbNegEntry )
											 
{
	ErrorNumber       err = kErrNoErr;
	unsigned long            x, y, ylimit;
	float             posDist, negDist, prob;
	// float             min;
	// int               minInd, i, j;
	int j;
	
	if ( (m_bProb == FALSE) || (nbNegEntry<=0) || (nbPosEntry<=0) )
    {
		return kErrMemFull;
    }
	
	// build the histogram
	for (x = 0; x < m_Reso; x++)
    {
		ylimit = m_Reso-x;
		
		for (y = 0; y < ylimit; y++)
		{
			posDist = 0;
			negDist = 0;
			for (j = nbPosEntry-1; j >= 0; j--)
			{
				// Compute mininum distance to all the models, divided
				// by the weight->greater weight smaller distance
				// and it needs to be LESS than the MAXDIST provided
				posDist += 
					(posGauss[j]).getProbability ((TVGAUSSTYPE)x/m_Reso,
					(TVGAUSSTYPE)y/m_Reso)*posW[j];
			}
			if ( posDist > 0 )
				posDist = (float) log(posDist);
			for (j = nbNegEntry-1; j >= 0; j--)
			{
				// Compute mininum distance to all the models, divided
				// by the weight->greater weight smaller distance
				// and it needs to be LESS than the MAXDIST provided
				negDist += 
					(negGauss[j]).getProbability ((TVGAUSSTYPE)x/m_Reso,
					(TVGAUSSTYPE)y/m_Reso)*
					negW[j];
			} // for j
			if ( negDist > 0 )
			{
				negDist = (float) log(negDist);
			}
			// if minimum distance is to a positive gaussian, then assign
			// 255, if not, assign 0
			//	  fValues[x*m_Reso+y] = ( (posDist > negDist) && (posDist < maxPosDist )) ? 255 : 0;
			prob = posDist-negDist;
			
			//prob = ( (prob>0)&&(prob>maxPosDist) ? prob : 0.0 );
			prob = (float) ( (prob>0) ? prob : 0.0 );
			m_pProbValues[x*m_Reso+y]=prob;
			//fValues[x*m_Reso+y] = ( (posDist > negDist) && (posDist > maxPosDist) ) ? 255 : 0;
			m_pValues[x*m_Reso+y] = ((prob>0.0) && (posDist>maxPosDist)) ? (unsigned char) (255*prob) : 0;
			//fValues[x*m_Reso+y] = (prob>0.0) ? 255*prob : 0.0;
		} // for y
    } // for x
	return err;
}


//
//    BuildFromStats
//

ErrorNumber THistogram::BuildFromStats (float maxDist, BYTE threshold)
{
	ErrorNumber		err = kErrNoErr;
	DWORD				x, y, ylimit;
	float				dist;
	
	// build the histogram
	if (threshold == 0)
		for (x = 0; x < m_Reso; x++)
		{
			ylimit = m_Reso-x;
			
			for (y = 0; y < ylimit; y++)
			{
				dist = m_pColorModel->getMahalanobisSquare ((TVGAUSSTYPE)x/m_Reso,
					(TVGAUSSTYPE)y/m_Reso);
				if (dist > maxDist)
					m_pValues[x*m_Reso+y] = 0;
				else
					m_pValues[x*m_Reso+y] = (BYTE)(255 - dist * 255/maxDist);
			}
		}
		else
			for (x = 0; x < m_Reso; x++)
			{
				ylimit = m_Reso-x;
				
				for (y = 0; y < ylimit; y++)
				{
					dist = m_pColorModel->getMahalanobisSquare ((TVGAUSSTYPE)x/m_Reso,
						(TVGAUSSTYPE)y/m_Reso);
					if (dist > maxDist)
						m_pValues[x*m_Reso+y] = 0;
					else
						m_pValues[x*m_Reso+y] = ((255 - dist * 255/maxDist) >= threshold ?
						255: 0);
				}
			}
			
			return err;
}




//
//    BuildGaussian
//

ErrorNumber THistogram::BuildGaussian  (TBitMap*	bitmap, 
										TBitMap*	support,
                                        LPVOID		supportColor,
                                        TGaussianf& gaussian, 
										CRect *bitmapRect,
										CRect *supportRect)
{
	ErrorNumber		err = kErrNoErr;
	CRect				sRect   = (bitmapRect != NULL) ? *bitmapRect : bitmap->GetValidRect ();
	CRect				vRect   = (supportRect != NULL) ? *supportRect : support->GetValidRect ();
	DWORD				sWidth  = abs(sRect.Width()); // It can be negative in a CRect
	DWORD				vWidth  = abs(vRect.Width()); // It can be negative in a CRect
	DWORD				sHeight = abs(sRect.Height());// It can be negative in a CRect
	DWORD				vHeight = abs(vRect.Height());// It can be negative in a CRect;
	DWORD				sPixelSize  = bitmap->PixelSize()/8;
	DWORD				vPixelSize  = support->PixelSize()/8;
	DWORD				sLineOffset = bitmap->LineOffset();
	DWORD				vLineOffset = support->LineOffset();
	
	DWORD				pixelValue = *((LPDWORD)supportColor);
	DWORD				r, g, b, lum;
	
	BYTE				*slptr, *spptr, *vlptr, *vpptr;
	DWORD				x, y;
	
	
	if (sRect != vRect ) {
		// source and destination have not the same size
		err = kErrWrongSize;
	} else {
		if ((sPixelSize != 4) || (vPixelSize != 4))
			// such pixel encoding can not be processed
			err = kErrCantHandleEncoding;
		else
		{
			// initialise the model
			gaussian.setZero ();
			
			// get the beginning of valid data
			slptr = (LPBYTE)(bitmap->GetBuffer ()) + sRect.top * sLineOffset +
				sRect.left * sPixelSize;
			vlptr = (LPBYTE)(support->GetBuffer ()) + vRect.top * vLineOffset +
				vRect.left * vPixelSize;
			
			// scan for lines
			for (y = 0; y < vHeight; y++)
			{
				// initialise pixel pointers
				spptr = slptr;
				vpptr = vlptr;
				
				// scan for pixels
				for (x = 0; x < vWidth; x++)
				{
					if (*((LPDWORD)vpptr) == pixelValue)
					{
						
						// get components value:BGRX
						b = *spptr;
						spptr++;
						g = *spptr;
						spptr++;
						r = *spptr;
						spptr+=2;
						
						lum = r + g + b;
						
						if (lum)
							gaussian.addSample (	(TVGAUSSTYPE)r/lum,
							(TVGAUSSTYPE)g/lum);
						else
							gaussian.addSample ((TVGAUSSTYPE)0, (TVGAUSSTYPE)0);
					}
					else
						spptr += sPixelSize;
					
					vpptr += vPixelSize;
				}
				
				// jump to next line
				slptr += sLineOffset;
				vlptr += vLineOffset;
			} // for y
			gaussian.getGaussianInv();
			
		} // else
	}	// else
	return err;
	
}